import os
import sys
import tempfile
from typing import Any, Dict

from fakeredis import FakeStrictRedis
import pytest

from chaosplt_scheduler.service import Services
from chaosplt_scheduler.scheduler import create_scheduler_queue, \
    create_scheduler, start_scheduler, stop_scheduler, ChaosPlatformScheduler


sys.path.insert(
    0, os.path.join(
        os.path.dirname(__file__)))


@pytest.fixture
def config() -> Dict[str, Any]:
    return {
        "debug": False,
        "grpc": {
            "address": "127.0.0.1:50980"
        },
        "redis": {
            "host": "127.0.0.1",
            "port": 6379,
            "queue": "chaosplatform"
        },
        "worker": {
            "debug": False,
            "queue_name": "fakequeue",
            "worker_name": "worker-test",
            "worker_directory": tempfile.tempdir,
            "add_random_suffix_to_worker_name": True,
            "supervising_interval": 0.1,
            "redis": {
                "host": "127.0.0.1",
                "port": 6379
            }
        }
    }


@pytest.fixture
def scheduler(config: Dict[str, Any]) -> ChaosPlatformScheduler:
    queue = create_scheduler_queue(
        "myqueue", FakeStrictRedis(singleton=False), is_async=False)
    sched = create_scheduler(queue, config)
    sched.interval = 1
    return sched


@pytest.fixture
def services() -> Services:
    return Services()



from typing import Any, Dict
from unittest.mock import ANY, MagicMock, patch

from chaosplt_scheduler.service import Services
from chaosplt_scheduler.server import initialize_all, release_all


@patch('chaosplt_scheduler.service.initialize_scheduler_queue', autospec=True)
@patch('chaosplt_scheduler.server.create_grpc_server', autospec=True)
def test_initialize_all(create_grpc_server, initialize_scheduler_queue,
                        config: Dict[str, Any], services: Services):

    scheduler = MagicMock()
    initialize_scheduler_queue.return_value = scheduler

    grpc_server = MagicMock()
    create_grpc_server.return_value = grpc_server

    try:
        services, grpc_s = initialize_all(config, services, None)

        assert services.scheduler is not None
        assert services.worker is not None
        assert grpc_s is not None
        assert grpc_s == grpc_server
        grpc_s.start.assert_called()
        scheduler.start.assert_called()
    finally:
        release_all(services, grpc_server)
        grpc_s.stop.assert_called_with(grace=1)
        scheduler.stop.assert_called()



# -*- coding: utf-8 -*-
import glob
import os

import cherrypy
from dotenv import load_dotenv

__all__ = ["load_settings"]


def load_settings(env_path: str):
    """
    Load settings from the environment:

    * if `env_path` is a file, read it
    * if `env_path` is a directory, load, all its `*.env` files
    """
    if os.path.isdir(env_path):
        pattern = os.path.join(env_path, '**', '.env')
        for env_file in glob.iglob(pattern, recursive=True):
            load_dotenv(dotenv_path=env_file)
    else:
        load_dotenv(dotenv_path=env_path)

    debug = True if os.getenv('CHAOSPLATFORM_DEBUG') else False
    cherrypy.engine.unsubscribe('graceful', cherrypy.log.reopen_files)
    cherrypy.config.update({
        'engine.autoreload.on': False,
        'log.screen': debug,
        'log.access_file': '',
        'log.error_file': '',
        'environment': '' if debug else 'production'
    })

    config = {
        "debug": debug,
        "grpc": {
            "address": os.getenv("GRPC_LISTEN_ADDR"),
        },
        "queue_name": os.getenv("QUEUE_NAME"),
        "worker_name": os.getenv("WORKER_NAME"),
        "add_random_suffix_to_worker_name": True,
        "redis": {
            "host": os.getenv("REDIS_HOST", "127.0.0.1"),
            "port": os.getenv("REDIS_PORT", 6379)
        },
        "job": {
            "platform_url": os.getenv("PLATFORM_URL")
        }
    }

    return config



import logging
import logging.config
import pkgutil
from typing import Any, Dict

import simplejson as json

__all__ = ["clean_logger", "configure_logger"]
logger = logging.getLogger("chaosplatform")


def configure_logger(log_conf_path: str, config: Dict[str, Any]):
    if log_conf_path:
        with open(log_conf_path) as f:
            log_conf = json.load(f)
    else:
        log_conf = json.loads(
            pkgutil.get_data('chaosplt_scheduler', 'logging.json'))
    logging.config.dictConfig(log_conf)
    if config["debug"]:
        [h.setLevel(logging.DEBUG) for h in logger.handlers]
        logger.setLevel(logging.DEBUG)
    logger.debug("Logger configured")


def clean_logger():
    """
    Clean all handlers attached to the logger
    """
    for h in logger.handlers[:]:
        logger.removeHandler(h)



import logging
from typing import Any, Dict

from chaosplt_grpc.scheduler.server import \
    SchedulerService as GRPCSchedulerService

from .service import SchedulerService
from .scheduler import ChaosPlatformScheduler

__all__ = ["SchedulerRPC"]
logger = logging.getLogger("chaosplatform")


class SchedulerRPC(SchedulerService, GRPCSchedulerService):
    def __init__(self, scheduler: ChaosPlatformScheduler,
                 config: Dict[str, Any]):
        SchedulerService.__init__(self, scheduler, config)
        GRPCSchedulerService.__init__(self)



import time
from typing import Any, Dict

from fakeredis import FakeStrictRedis

from chaosplt_scheduler.scheduler import create_scheduler_queue, \
    create_scheduler, start_scheduler, stop_scheduler

from fixtures.jobs import count_words


def test_create_scheduler_queue():
    queue = create_scheduler_queue(
        "myqueue", FakeStrictRedis(singleton=False), is_async=False)
    assert queue is not None


def test_create_scheduler(config: Dict[str, Any]):
    queue = create_scheduler_queue(
        "myqueue", FakeStrictRedis(singleton=False), is_async=False)
    sched = create_scheduler(queue, config)
    assert sched is not None

    job = queue.enqueue(count_words, "hello world")
    assert job.is_finished
    assert job.result == 2


def test_running_scheduler(config: Dict[str, Any]):
    queue = create_scheduler_queue(
        "myqueue", FakeStrictRedis(singleton=False), is_async=False)
    sched = create_scheduler(queue, config)
    sched.interval = 1
    assert sched.running == False

    start_scheduler(sched)
    time.sleep(0.3)
    assert sched.running == True

    stop_scheduler(sched)
    time.sleep(0.3)
    assert sched.running == False



#!/usr/bin/env python
import sys

import click
from redis import Redis
from redis.exceptions import ConnectionError
from rq import Queue, Worker
from rq.cli.helpers import setup_loghandlers_from_args
from rq.contrib.legacy import cleanup_ghosts
from rq.job import Job

from . import __version__


@click.group()
@click.version_option(version=__version__)
def runner():
    pass


@runner.command()
@click.option('--name', help='Worker name.', required=True)
@click.option('--queue', help='Queue name.', required=True)
@click.option('--verbose', is_flag=True, help='Enable more traces.')
@click.option('--connection-host', help='Connection host', default='localhost',
              show_default=True)
@click.option('--connection-port', help='Connection port', default=6379,
              show_default=True, type=int)
@click.option('--connection-password', help='Connection password')
@click.option('--connection-db', help='Connection database to use', default=0,
              show_default=True)
@click.option('--connection-over-tls', is_flag=True,
              help='Connection over TLS')
def run(name: str, queue: str, connection_host: str,
        connection_port: int = 6379, connection_password: str = None,
        connection_over_tls: bool = False, connection_db: int = 0,
        verbose: bool = False):
    date_format = "%Y-%m-%dT%H:%M:%S"
    log_format = "%(asctime)s.%(msecs)03d [%(levelname)s] %(message)s"
    setup_loghandlers_from_args(verbose, False, date_format, log_format)

    params = {
        "ssl": connection_over_tls,
        "password": connection_password,
        "host": connection_host,
        "port": connection_port,
        "db": connection_db
    }

    try:
        c = Redis(**params)
        cleanup_ghosts(c)
        worker = Worker(
            [queue],
            name=name,
            connection=c,
            default_worker_ttl=420,
            default_result_ttl=500,
            job_monitoring_interval=30,
            job_class=Job,
            queue_class=Queue,
            exception_handlers=None
        )

        worker.work(
            burst=False, logging_level="DEBUG" if verbose else "INFO",
            date_format=date_format, log_format=log_format)
    except ConnectionError as e:
        print(e)
        sys.exit(1)


if __name__ == '__main__':
    run()



import logging
from typing import Any, Dict, NoReturn

import attr
from dateparser import parse

from .runner import run as run_toolkit
from .scheduler import initialize_scheduler_queue, start_scheduler, \
    stop_scheduler, ChaosPlatformScheduler
from .worker import ChaosPlatformWorker, initialize_worker, start_worker, \
    stop_worker

__all__ = ["SchedulerService", "initialize_services", "shutdown_services"]
logger = logging.getLogger("chaosplatform")


@attr.s
class Services:
    scheduler: ChaosPlatformScheduler = attr.ib(default=None)
    worker: ChaosPlatformWorker = attr.ib(default=None)


class SchedulerService:
    def __init__(self, scheduler: ChaosPlatformScheduler,
                 config: Dict[str, Any]):
        self.scheduler = scheduler
        self.job_config = config.get("job")

    def release(self):
        self.scheduler = None
        self.job_config = None

    def schedule(self, schedule_id: str, user_id: str, org_id: str,
                 workspace_id: str, experiment_id: str, token_id: str,
                 token: str, scheduled: str, experiment: str, interval: int,
                 repeat: int, cron: str, settings: str, configuration: str,
                 secrets: str, timezone: str = 'UTC') -> str:
        """
        Queue a job for execution

        Various scenarios can happen:

        * When `cron` is set to a Cron string, the job will be queued to
          run periodically.
        * When `interval` is set to an integer, the job is queued to run
          periodically
        * Otherwise, the job is scheduled to run once

        However, when `repeat` is also set, the job is ran only that number of
        times, when `cron` or `interval` were also set.

        The job is first queued when `scheduled` is reached. It must be a
        datestring in the given `timezone`, which defaults to `"UTC"`.

        The method returns the job identifier provided by the underlying
        scheduler.
        """
        # None means infinity
        repeat = repeat or None
        scheduled = parse(scheduled, settings={'TO_TIMEZONE': timezone})
        platform_url = self.job_config["platform_url"]
        if cron:
            job = self.scheduler.cron(
                cron,
                func=run_toolkit,
                args=[
                    schedule_id, user_id, org_id,
                    workspace_id, experiment_id, token_id, token,
                    experiment, platform_url, settings, configuration, secrets
                ],
                repeat=repeat
            )
        elif interval:
            job = self.scheduler.schedule(
                scheduled,
                func=run_toolkit,
                args=[
                    schedule_id, user_id, org_id,
                    workspace_id, experiment_id, token_id, token,
                    experiment, platform_url, settings, configuration, secrets
                ],
                interval=interval,
                repeat=repeat
            )
        else:
            job = self.scheduler.enqueue_at(
                scheduled,
                run_toolkit,
                schedule_id, user_id, org_id, workspace_id, experiment_id,
                token_id, token, experiment, platform_url, settings,
                configuration, secrets
            )
        logger.info("Scheduled job {} at '{}Z'".format(
            job.id, scheduled.isoformat()))
        return job.id

    def cancel(self, job_id: str) -> NoReturn:
        """
        Cancel a job.

        If the job never started, it is cancelled beforehand. Or its next
        iteration, if one was planned.
        """
        logger.info("Cancel job {}".format(job_id))
        self.scheduler.cancel(job_id)


def initialize_services(services: Services, config: Dict[str, Any]):
    """
    Initialize the scheduler and worker resources
    """
    if not services.scheduler:
        scheduler = initialize_scheduler_queue(config)
        start_scheduler(scheduler)
        services.scheduler = SchedulerService(scheduler, config)

    if not services.worker:
        worker_config = config["worker"]
        services.worker = initialize_worker(
            worker_config["queue_name"], worker_config["worker_name"], config)
        start_worker(services.worker)


def shutdown_services(services: Services):
    """
    Shutdown the scheduler and worker services, releasing their resources
    """
    if services.scheduler:
        sched = services.scheduler.scheduler
        stop_scheduler(sched)
        services.scheduler.release()
    if services.worker:
        stop_worker(services.worker)



import logging
import shutil
import threading
import time
from typing import Any, Dict
import uuid

import psutil
from psutil import Popen

__all__ = ["initialize_worker", "ChaosPlatformWorker", "start_worker",
           "stop_worker"]
logger = logging.getLogger("chaosplatform")


class ChaosPlatformWorker:
    def __init__(self, queue: str, worker_name: str, worker_directory: str,
                 log_level: str = "INFO", count_worker: int = 1,
                 add_suffix: bool = False, supervising_interval: int = 2):
        # worker params
        self.queue_name = queue
        self.worker_name = worker_name
        self.worker_directory = worker_directory
        self.log_level = log_level
        self.count_worker = count_worker
        self.add_suffix = add_suffix

        # internal machinery
        self._lock = threading.Lock()
        self._running = False
        self._processes = []
        self.supervising_interval = supervising_interval
        self.supervisor = threading.Thread(None, self.supervise)

    @property
    def running(self) -> bool:
        with self._lock:
            return self._running

    @running.setter
    def running(self, value: bool):
        with self._lock:
            self._running = value

    @property
    def workers_count(self):
        return len(self._processes)

    def start(self):
        """
        Start the worker and starts supervising its processes
        """
        logger.info('Starting {} workers...'.format(self.count_worker))
        self.running = True
        list(map(self._start_worker, range(self.count_worker)))
        self.supervisor.start()

    def stop(self):
        """
        Terminate all the process workers
        """
        logger.info('Stopping all workers...')
        self.running = False
        for p in self._processes:
            logger.info("Terminating worker (PID {})".format(p.pid))
            p.terminate()
            ret = p.wait()
            logger.info(
                "Worker (PID {}) terminated with return code {}".format(
                    p.pid, ret))
        self.supervisor.join(timeout=0.5)
        self._processes = []

    def supervise(self):
        """
        Supervise all workers and restart any instance that has died
        """
        logger.info("Starting worker supervisors...")
        while self.running:
            for i, p in enumerate(self._processes[:]):
                # exit fast
                if not self.running:
                    return

                if not p.is_running() or p.status() == psutil.STATUS_ZOMBIE:
                    self._processes.remove(p)
                    logger.warning(
                        "Worker (PID {}) gone, starting a new one...".format(
                            p.pid))
                    self._start_worker(i)

            time.sleep(self.supervising_interval)
        logger.info("Worker supervisor is now terminated")

    def _start_worker(self, index: int):
        """
        Start a new worker child process
        """
        logger.info("Creating a new worker")

        path = shutil.which("chaosplatform-worker")
        name = self.worker_name
        if self.add_suffix:
            name = "{}-{}".format(name, uuid.uuid4().hex)

        args = [
            path,
            "--name", name,
            "--queue", self.queue_name
        ]

        if self.log_level == "DEBUG":
            args.append("--verbose")

        # notice, the current's process environment variables will be passed
        # down to the child worker process
        logger.debug("Executing: {}".format(" ".join(args)))
        p = Popen(
            args, stdout=None,
            stderr=None, shell=False, cwd=self.worker_directory)
        self._processes.append(p)
        logger.info("Worker '{}' started (PID {})".format(name, p.pid))


def initialize_worker(queue_name: str, worker_name: str,
                      config: Dict[str, Any]) -> ChaosPlatformWorker:
    """
    Create a worker instance and attach a connection to a redis server
    """
    worker_config = config["worker"]
    worker = ChaosPlatformWorker(
        queue=queue_name, worker_name=worker_name,
        add_suffix=worker_config.get("add_random_suffix_to_worker_name"),
        worker_directory=worker_config["worker_directory"],
        log_level="DEBUG" if worker_config["debug"] else "INFO",
        count_worker=int(worker_config.get("count", 1)),
        supervising_interval=float(worker_config.get(
            "supervising_interval", 2))
    )
    return worker


def start_worker(worker: ChaosPlatformWorker):
    """
    Run the worker
    """
    logger.info("Starting worker")
    worker.start()


def stop_worker(worker: ChaosPlatformWorker):
    """
    Stop and release the worker
    """
    logger.info("Stopping worker")
    worker.stop()



import logging
import threading
import time
from typing import Any, Dict

from redis import Redis
from rq import Queue
from rq.job import Job
from rq_scheduler import Scheduler
from rq_scheduler.utils import setup_loghandlers

__all__ = ["create_scheduler", "ChaosPlatformScheduler", "start_scheduler",
           "stop_scheduler", "create_scheduler_queue"]
logger = logging.getLogger("chaosplatform")


class ChaosPlatformScheduler(threading.Thread, Scheduler):
    def __init__(self, *args, **kwargs):
        """
        A scheduler based on the `rq` scheduler but in its own thread.
        """
        threading.Thread.__init__(self)
        Scheduler.__init__(self, *args, interval=10, **kwargs)

        self._lock = threading.Lock()
        self._running = False

    @property
    def running(self):
        """
        Flag indicating whether the scheduler is currently active
        """
        with self._lock:
            return self._running

    @running.setter
    def running(self, state: bool):
        """
        Mark the scheduler as active or inactive
        """
        with self._lock:
            self._running = state

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, step: int):
        self._interval = step

    def stop(self):
        """
        Terminate the scheduler

        When done, the current thread is exited.
        """
        self.log.info('Stopping scheduler...')
        self.running = False
        self.register_death()
        self.remove_lock()

    def run(self):
        """
        Scheduler thread mainloop

        Runs until the `running` flag is set to `False`.
        """
        self.log.info('Starting scheduler...')
        self.running = True
        self.register_birth()

        while self.running:
            start_time = time.time()
            if self.acquire_lock():
                self.enqueue_jobs()
                self.remove_lock()

            # from rq-scheduler run
            seconds_elapsed_since_start = time.time() - start_time
            seconds_until_next_scheduled_run = \
                self._interval - seconds_elapsed_since_start
            if self.running and seconds_until_next_scheduled_run > 0:
                time.sleep(seconds_until_next_scheduled_run)

        self.log.info('Exiting scheduler...')


def create_scheduler(queue: Queue, config: Dict[str, Any],
                     job_class: Job = Job) -> ChaosPlatformScheduler:
    """
    Create a scheduler on the given queue and job class
    """
    log_level = 'DEBUG' if config.get('debug') else 'INFO'
    setup_loghandlers(level=log_level)
    return ChaosPlatformScheduler(
        queue=queue, connection=queue.connection,
        queue_class=queue.__class__,
        job_class=job_class)


def create_scheduler_queue(name: str, connection: Redis,
                           is_async: bool = True) -> Queue:
    """
    Instanciate the scheduler queue connection
    """
    return Queue(name, connection=connection, is_async=is_async)


def start_scheduler(scheduler: ChaosPlatformScheduler):
    """
    Run the scheduler application forever
    """
    logger.info("Starting scheduler")
    scheduler.start()


def stop_scheduler(scheduler: ChaosPlatformScheduler):
    """
    Stop and release the scheduler
    """
    logger.info("Stopping scheduler")
    scheduler.stop()


def initialize_scheduler_queue(config: Dict[str, Any]) \
                               -> ChaosPlatformScheduler:
    queue_name = config["redis"].pop("queue")
    connection_params = config.get("redis")
    connection = Redis(**connection_params)
    queue = create_scheduler_queue(queue_name, connection=connection)
    return create_scheduler(queue=queue, config=config)



import logging
from typing import Any, Dict, Tuple

from chaosplt_grpc import create_grpc_server, start_grpc_server, \
    stop_grpc_server
from chaosplt_grpc.scheduler.server import register_scheduler_service
import cherrypy
from grpc import Server

from .rpc import SchedulerRPC
from .scheduler import ChaosPlatformScheduler
from .service import Services, initialize_services, shutdown_services

__all__ = ["initialize_all", "release_all", "run_forever"]
logger = logging.getLogger("chaosplatform")


def initialize_all(config: Dict[str, Any], services: Services = None,
                   grpc_server: Server = None) -> Tuple[Services, Server]:
    """
    Initialize all resources for the service:

    * The gRPC application
    * The scheduler thread
    """
    logger.info("Initializing scheduler service resources")
    if not services:
        services = Services()

    initialize_services(services, config)
    grpc_server = initialize_grpc(
        services.scheduler.scheduler, config, grpc_server)

    return (services, grpc_server)


def release_all(services: Services, grpc_server: Server):
    """
    Release all resources
    """
    logger.info("Releasing scheduler service resources")
    if grpc_server:
        logger.info("gRPC server stopping")
        stop_grpc_server(grpc_server)
        logger.info("gRPC server stopped")
    shutdown_services(services)


def run_forever(config: Dict[str, Any]):
    """
    Run and block until a signal is sent to the process.

    The application, services or gRPC server are all created and initialized
    when the application starts.
    """
    def run_stuff(config: Dict[str, Any]):
        resources = initialize_all(config)
        cherrypy.engine.subscribe(
            'stop', lambda: release_all(*resources),
            priority=20)

    cherrypy.engine.subscribe(
        'start', lambda: run_stuff(config), priority=60)

    cherrypy.server.unsubscribe()
    cherrypy.engine.signals.subscribe()
    cherrypy.engine.start()
    cherrypy.engine.block()


###############################################################################
# Internals
###############################################################################
def initialize_grpc(scheduler: ChaosPlatformScheduler, config: Dict[str, Any],
                    grpc_server: Server = None) -> Server:
    """
    Initialize the gRPC server

    Create the server if not provided. This called by `initialize_all`.
    """
    if not grpc_server:
        srv_addr = config["grpc"]["address"]
        if srv_addr:
            grpc_server = create_grpc_server(srv_addr)
            start_grpc_server(grpc_server)
            logger.info("gRPC server started on {}".format(srv_addr))

        register_scheduler_service(
            SchedulerRPC(scheduler, config), grpc_server)

    return grpc_server



from datetime import datetime, timedelta
from unittest.mock import MagicMock, ANY

from dateparser import parse

from chaosplt_scheduler.rpc import SchedulerRPC


def test_schedule_experiment_execution_at():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.enqueue_at.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()
    scheduled = parse(dt, settings={'TO_TIMEZONE': 'UTC'})

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=None,
        repeat=None,
        cron=None,
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.enqueue_at.assert_called_with(
        scheduled, ANY, "1", "2", "3", "4", "5", "6", "7", "8", 
        "http://a:80", None, None, None
    )
    assert job_id == "myjob"


def test_schedule_experiment_execution_cron_no_repeat():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.cron.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=None,
        repeat=None,
        cron="1 * * * *",
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.cron.assert_called_with(
        "1 * * * *", func=ANY,
        args=[
            "1", "2", "3", "4", "5", "6", "7", "8", "http://a:80", None, None,
            None
        ],
        repeat=None
    )


def test_schedule_experiment_execution_cron_with_repeat():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.cron.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=None,
        repeat=10,
        cron="1 * * * *",
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.cron.assert_called_with(
        "1 * * * *", func=ANY,
        args=[
            "1", "2", "3", "4", "5", "6", "7", "8", "http://a:80", None, None,
            None
        ],
        repeat=10
    )


def test_schedule_experiment_execution_interval_no_repeat():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.schedule.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()
    scheduled = parse(dt, settings={'TO_TIMEZONE': 'UTC'})

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=30,
        repeat=None,
        cron=None,
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.schedule.assert_called_with(
        scheduled, func=ANY,
        args=[
            "1", "2", "3", "4", "5", "6", "7", "8", "http://a:80", None, None,
            None
        ],
        interval=30, repeat=None
    )


def test_schedule_experiment_execution_interval_no_repeat():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.schedule.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()
    scheduled = parse(dt, settings={'TO_TIMEZONE': 'UTC'})

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=30,
        repeat=10,
        cron=None,
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.schedule.assert_called_with(
        scheduled, func=ANY,
        args=[
            "1", "2", "3", "4", "5", "6", "7", "8", "http://a:80", None, None,
            None
        ],
        interval=30, repeat=10
    )


def test_cancel_schedule():
    scheduler = MagicMock()
    job = MagicMock()
    job.id = "myjob"
    scheduler.enqueue_at.return_value = job

    dt = (datetime.utcnow() + timedelta(seconds=1)).isoformat()
    scheduled = parse(dt, settings={'TO_TIMEZONE': 'UTC'})

    rpc = SchedulerRPC(scheduler, {"job": {"platform_url": "http://a:80"}})
    job_id = rpc.schedule(
        schedule_id="1",
        user_id="2",
        org_id="3",
        workspace_id="4",
        experiment_id="5",
        token_id="6",
        token="7",
        scheduled=dt,
        experiment="8",
        interval=None,
        repeat=None,
        cron=None,
        settings=None,
        configuration=None,
        secrets=None)
    scheduler.enqueue_at.assert_called_with(
        scheduled, ANY, "1", "2", "3", "4", "5", "6", "7", "8", "http://a:80",
        None, None, None
    )
    assert job_id == "myjob"

    rpc.cancel(job_id)
    scheduler.cancel.assert_called_with(job_id)



import tempfile
import time
from typing import Any, Dict
from unittest.mock import ANY, MagicMock, patch

from psutil import Popen
from rq import Queue

from chaosplt_scheduler.worker import initialize_worker


@patch("chaosplt_scheduler.worker.Popen", autospec=True)
def test_initialize_worker(popen: Popen, config: Dict[str, Any]):
    worker_config = config.get("worker")
    queue_name = worker_config.get("queue_name")
    worker_name = worker_config.get("worker_name")
    worker = initialize_worker(
        queue_name,
        worker_name,
        config
    )
    assert worker is not None
    assert worker.running is False
    assert worker.workers_count == 0

    try:
        worker.start()
        assert worker.running is True
        assert worker.workers_count == 1
        popen.assert_called_with(
            ANY, stdout=None, stderr=None, shell=False, cwd=tempfile.tempdir
        )
    finally:
        worker.stop()
        assert worker.running is False
        assert worker.workers_count == 0


@patch("chaosplt_scheduler.worker.Popen", autospec=True)
def test_failed_worker_is_replaced(popen: Popen, config: Dict[str, Any]):
    worker_config = config.get("worker")
    queue_name = worker_config.get("queue_name")
    worker_name = worker_config.get("worker_name")
    worker = initialize_worker(
        queue_name,
        worker_name,
        config
    )
    assert worker is not None
    assert worker.running is False
    assert worker.workers_count == 0

    proc1 = MagicMock()
    proc2 = MagicMock()

    popen.side_effect = [proc1, proc2]
    proc1.is_running.return_value = True
    proc2.is_running.return_value = True

    try:
        worker.start()
        assert worker.running is True
        assert worker.workers_count == 1
        popen.assert_called_with(
            ANY, stdout=None, stderr=None, shell=False, cwd=tempfile.tempdir
        )
        proc1.is_running.return_value = False
        time.sleep(0.2)
        assert worker.workers_count == 1
        assert worker._processes[0] == proc2
    finally:
        worker.stop()
        assert worker.running is False
        assert worker.workers_count == 0



import logging

from chaoslib.experiment import run_experiment
from chaoslib.types import Journal
import requests
import simplejson as json

__all__ = ["run"]
logger = logging.getLogger("chaosplatform")


def run(schedule_id: str, user_id: str, org_id: str, workspace_id: str,
        experiment_id: str, token_id: str, token: str, experiment: str,
        platform_url: str, settings: str = None, configuration: str = None,
        secrets: str = None) -> Journal:
    """
    Execute an experiment and return its journal

    See also:
    https://docs.chaostoolkit.org/reference/api/journal/
    """
    settings = json.loads(settings) if settings else {}
    experiment = json.loads(experiment)

    if configuration:
        if not isinstance(configuration, dict):
            configuration = json.loads(configuration)
        experiment.setdefault("configuration", {})
        experiment["configuration"].update(configuration)

    if secrets:
        if not isinstance(secrets, dict):
            secrets = json.loads(secrets)
        experiment.setdefault("secrets", {})
        experiment["secrets"].update(secrets)

    try:
        journal = run_experiment(experiment)
    except Exception as x:
        logger.error(
            "Running experiment failed due to {}".format(str(x)),
            exc_info=True
        )
    else:
        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer {}".format(token),
            "Connection": "close"
        }
        r = requests.post(
            "{}/api/v1/executions".format(platform_url), json={
                "experiment_id": experiment_id,
                "journal": journal
            }, headers=headers)

        if r.status_code > 399:
            logger.error(
                "Failed to push journal results for experiment {}: {}".format(
                    experiment_id, r.text
                ))

        return journal
